package net.knarcraft.ffmpegconverter.converter;

import net.knarcraft.ffmpegconverter.streams.AudioStream;
import net.knarcraft.ffmpegconverter.streams.StreamObject;
import net.knarcraft.ffmpegconverter.streams.VideoStream;

import java.io.File;
import java.io.IOException;
import java.util.List;

public class VideoConverter extends Converter {
    private String newExt;

    /**
     * Instantiates a new video converter
     * @param ffprobePath <p>Path/command to ffprobe.</p>
     * @param ffmpegPath <p>Path/command to ffmpeg.</p>
     * @param newExt <p>The extension of the new file.</p>
     */
    public VideoConverter(String ffprobePath, String ffmpegPath, String newExt) {
        this.ffprobePath = ffprobePath;
        this.ffmpegPath = ffmpegPath;
        this.newExt = newExt;
    }

    /**
     * Reads streams from a file, and converts it to an mp4
     * @param folder <p>The folder of the file to process.</p>
     * @param file <p>The file to process.</p>
     * @throws IOException <p>If the BufferedReader fails.</p>
     */
    private void processFile(File folder, File file, String newExt) throws IOException {
        List<StreamObject> streams = probeFile(ffprobePath, file);
        if (streams.size() == 0) {
            throw new IllegalArgumentException("The file has no streams");
        }
        String newPath = fileCollisionPrevention(folder.getAbsolutePath() + File.separator + stripExtension(file) + "." + newExt, newExt);
        convertProcess(new ProcessBuilder(builderCommand(ffmpegPath, file.getName(), streams, newPath, folder)), folder);
    }

    /**
     * Generates a command for a ProcessBuilder
     * @param executable <p>The executable file for ffmpeg.</p>
     * @param fileName <p>The input file.</p>
     * @param streams <p>A list of ffprobe streams.</p>
     * @param outFile <p>The output file.</p>
     * @return <p>A list of commands</p>
     */
    private String[] builderCommand(String executable, String fileName, List<StreamObject> streams, String outFile, File folder) {
        List<String> command = generalFile(executable, fileName);

        if (this.DEBUG) {
            addDebug(command, 50, 120);
        }

        List<AudioStream> audioStreams = filterStreamsByType(streams, "audio");
        List<VideoStream> videoStreams = filterStreamsByType(streams, "video");

        VideoStream videoStream = null;
        AudioStream audioStream = null;
        if (videoStreams.size() > 0) {
            videoStream = videoStreams.get(0);
        }
        if (audioStreams.size() > 0) {
            audioStream = audioStreams.get(0);
        }

        boolean videoAdded = addSubtitles(command, folder, fileName, videoStream);

        if (!videoAdded && videoStreams.size() > 0) {
            command.add("-map");
            command.add("0:" + videoStream);
        }
        if (audioStreams.size() > 0) {
            command.add("-map");
            command.add("0:" + audioStream);
        }

        convertToStereo(command);

        command.add(outFile);
        return command.toArray(new String[0]);
    }

    /**
     * Adds subtitles to a command
     * @param command <p>The command to add to.</p>
     * @param folder <p>The folder containing the file to be converted.</p>
     * @param fileName <p>The name of the file to be converted.</p>
     * @param videoStream <p>The video stream to be added.</p>
     * @return <p>True if a video stream was added.</p>
     */
    private boolean addSubtitles(List<String> command, File folder, String fileName, VideoStream videoStream) {
        String externalSubtitle = hasExternalSubtitle(folder.getAbsolutePath(), fileName);
        String externalImageSubtitle = hasExternalImageSubtitle(folder.getAbsolutePath(), fileName);
        if (!externalSubtitle.equals("")) {
            command.add("-vf");
            command.add("subtitles=" + externalSubtitle);
        } else if (!externalImageSubtitle.equals("")) {
            command.add("-i");
            command.add(externalImageSubtitle);
            if (this.DEBUG) {
                addDebug(command, 50, 120);
            }
            //TODO: Scale subtitles to video
            command.add("-filter_complex");
            command.add("[1:s]scale=width=1920:height=800,crop=w=1920:h=800:x=0:y=out_h[sub];[" + videoStream +
                    ":v][sub]overlay");
            command.add("-profile:v");
            command.add("baseline");
            return true;
        }
        return false;
    }

    /**
     * Converts the audio of a video to stereo
     * @param command <p>The command list to add to.</p>
     */
    private void convertToStereo(List<String> command) {
        command.add("-af");
        command.add("pan=stereo|FL < 1.0*FL + 0.707*FC + 0.707*BL|FR < 1.0*FR + 0.707*FC + 0.707*BR");
    }

    @Override
    public String[] getValidFormats() {
        return VIDEO_FORMATS;
    }

    @Override
    public void convert(File file) throws IOException {
        processFile(file.getParentFile(), file, newExt);
    }
}